from typing import TYPE_CHECKING

import dash
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State

from .elem_ids import ElemIds, SidedId
from .style import (
    CONTENT_STYLE,
    CONTENT_STYLE_HIDDEN,
    DEFAULT_VIEW_NAME,
    LEFT_COLOR,
    RIGHT_COLOR,
    SIDEBAR_HIDDEN,
    SIDEBAR_STYLE,
    colors,
)
from .util import table_from_pandas

if TYPE_CHECKING:
    from .default_report import DefaultReport
    from .peeker_class import Peeker


def add_base_layout_to_app(
    app: dash.Dash, peeker: "Peeker", default: "DefaultReport"
):

    navbar = _get_navbar(peeker.view_store.get_all_ids())
    sidebar = _get_sidebar(peeker)

    content = html.Div(
        _get_report_layout(peeker, default),
        id=ElemIds.main_content,
        style=CONTENT_STYLE,
    )

    app.layout = html.Div(
        [
            dcc.Store(id=ElemIds.click_store),
            dcc.Location(id=ElemIds.url),
            navbar,
            sidebar,
            content,
        ],
    )
    add_sidebar_callback(app)


def add_sidebar_callback(app):
    @app.callback(
        [
            Output(ElemIds.sidebar, "style"),
            Output(ElemIds.main_content, "style"),
            Output(ElemIds.click_store, "data"),
        ],
        [Input(ElemIds.sidebar_button, "n_clicks")],
        [
            State(ElemIds.click_store, "data"),
        ],
    )
    def toggle_sidebar(n, nclick):
        if n:
            if nclick == "SHOW":
                sidebar_style = SIDEBAR_HIDDEN
                content_style = CONTENT_STYLE_HIDDEN
                cur_nclick = "HIDDEN"
            else:
                sidebar_style = SIDEBAR_STYLE
                content_style = CONTENT_STYLE
                cur_nclick = "SHOW"
        else:
            sidebar_style = SIDEBAR_STYLE
            content_style = CONTENT_STYLE
            cur_nclick = "SHOW"

        return sidebar_style, content_style, cur_nclick


def _get_navbar(all_views):

    return dbc.Navbar(
        children=html.Div(
            [
                html.Div(
                    dbc.Button(
                        "Toggle Filters",
                        color="secondary",
                        className="mr-1",
                        id=ElemIds.sidebar_button,
                    ),
                    className="col-md-2",
                ),
                *_get_dd("Left Column", ElemIds.left.dd, all_views),
                *_get_dd("Right Column", ElemIds.right.dd, all_views),
            ],
            className="row w-100",
        ),
        color="dark",
        dark=True,
        fixed="top",
    )


def _get_dd(label, id_, options):
    return [
        html.Div(
            html.H2(label, style={"color": "white"}),
            className="col-md-2",
            style={"padding-top": 6},
        ),
        html.Div(
            dcc.Dropdown(
                options=[{"value": x, "label": x} for x in options],
                id=id_,
            ),
            className="col-md-3",
        ),
    ]


def _get_sidebar(peeker):
    return html.Div(
        children=dbc.Tabs(
            [
                dbc.Tab(
                    children=_get_filter_tab(peeker, ind),
                    label=DEFAULT_VIEW_NAME,
                    label_style={
                        "color": color,
                        "font-weight": "bold",
                        "padding": 18,
                    },
                    id=side.tab,
                )
                for color, side, ind in zip(colors, ElemIds.sides(), range(2))
            ],
            className="w-100",
        ),
        id=ElemIds.sidebar,
        style=SIDEBAR_STYLE,
    )


def _get_filter_tab(peeker, ind):

    side_ids = ElemIds.sides()[ind]
    elems = _get_filter_tab_top(side_ids)

    for col in peeker.cat_filter_cols:
        options = peeker.default_values[col]
        filter_id = peeker.id_repo.get_filter_id(col, ind)
        elems.append(_get_cat_filter_elem(col, options, options, filter_id))

    for col in peeker.num_filter_cols:
        filter_id = peeker.id_repo.get_filter_id(col, ind)
        selected = peeker.default_values[col]
        binends = peeker.num_ranges[col]
        elems.append(_get_num_filter_elem(col, binends, selected, filter_id))

    elems.append(html.Div(className="row", style={"height": 100}))
    return html.Div(elems, className="container")


def _get_filter_tab_top(side_id: SidedId):
    return [
        html.Div(
            [
                html.Div(html.H4("View Name:"), className="col-md-3"),
                html.Div(
                    dcc.Input(
                        value=DEFAULT_VIEW_NAME,
                        id=side_id.name_input,
                        maxLength=23,
                    ),
                    className="col-md-9",
                ),
            ],
            className="row",
            style={"padding": 10},
        ),
        html.Div(
            [
                html.Div(html.H4("View Description:"), className="col-md-12"),
                html.Div(
                    dcc.Textarea(
                        className="w-100",
                        id=side_id.desc_input,
                    ),
                    className="col-md-12",
                ),
            ],
            className="row",
            style={"padding": 10},
        ),
        html.Div(
            [
                html.Div(
                    dbc.Button(
                        "Save View", color="success", id=side_id.save_button
                    ),
                    className="col-md-3",
                ),
                html.Div(
                    dbc.Button(
                        "Apply Filters",
                        color="primary",
                        id=side_id.apply_button,
                    ),
                    className="col-md-3",
                ),
                html.Div(
                    dbc.Button(
                        "Delete View", color="danger", id=side_id.del_button
                    ),
                    className="col-md-3",
                ),
            ],
            className="d-flex justify-content-center row",
        ),
        html.Hr(),
    ]


def _get_report_layout(peeker, default):
    # TODO make these categories renamable
    report_bases = [
        html.Div(
            html.Div(
                [
                    html.Div(
                        html.H2(name),
                        className="row w-100",
                        style={"margin": 35},
                    ),
                    html.Div(
                        children=fun(peeker, default),
                        className="d-flex justify-content-center row",
                    ),
                ],
                className="col-md-12",
            ),
            className="row",
        )
        for name, fun in zip(
            ["Metrics", "Shared Figures", "Standalone Figures", "Tables"],
            [
                _get_metrics_base,
                _get_shared_figs_base,
                _get_so_figs_base,
                _get_tables_base,
            ],
        )
    ]

    return html.Div(
        report_bases
        + [
            dcc.Store(id=side_id.saved_view_option_store)
            for side_id in ElemIds.sides()
        ],
        className="container-fluid",
    )


def _get_metrics_base(peeker: "Peeker", default: "DefaultReport"):
    table_rows = [
        html.Tr(
            [
                html.Th(),
                html.Th(
                    DEFAULT_VIEW_NAME,
                    style={"color": LEFT_COLOR},
                    id=ElemIds.left.metric_table_header,
                ),
                html.Th(
                    DEFAULT_VIEW_NAME,
                    style={"color": RIGHT_COLOR},
                    id=ElemIds.right.metric_table_header,
                ),
                html.Th("Diff"),
            ]
        )
    ]

    for m in default.default_metrics:
        v = m.value
        mid = peeker.id_repo.new_metric_id(m.name)
        table_rows.append(
            html.Tr(
                [
                    html.Td(html.B(m.name)),
                    html.Td(v, id=mid.left),
                    html.Td(v, id=mid.right),
                    html.Td(id=mid.diff),
                ]
            )
        )

    return html.Div(
        html.Table(table_rows),
        className="d-flex justify-content-center col-md-12",
    )


def _get_shared_figs_base(peeker: "Peeker", default: "DefaultReport"):
    fig_divs = []
    for fig_elem in default.shared_figures:
        trace = default.get_default_trace(fig_elem.name)
        trace.name = DEFAULT_VIEW_NAME
        trace.marker.color = LEFT_COLOR
        trace2 = type(trace)(trace)
        fig = fig_elem.fig
        fig.add_trace(trace)
        trace2.marker.color = RIGHT_COLOR
        fig.add_trace(trace2)
        fig_id = peeker.id_repo.new_shared_fig_id(fig_elem.name)
        fig_divs.append(
            html.Div(
                [
                    dcc.Graph(figure=fig, id=fig_id.id_),
                    dcc.Store(data=trace, id=fig_id.left),
                    dcc.Store(data=trace2, id=fig_id.right),
                ],
                className=f"col-md-{fig_elem.col_width}",
            )
        )

    return fig_divs


def _get_so_figs_base(peeker: "Peeker", default: "DefaultReport"):

    so_figs = []
    for i, fig_elem in enumerate(default.default_standalone_figures):
        fig_id = peeker.id_repo.new_so_fig_id(f"so-fig-{i}")
        for id_, color in zip(fig_id, colors):
            so_figs.append(
                html.Div(
                    dcc.Graph(figure=fig_elem.fig, id=id_),
                    className=f"col-md-{fig_elem.col_width}",
                    style={
                        "border-color": color,
                        "border-style": "solid",
                        "border-radius": "10px",
                        "padding": 4,
                    },
                )
            )

    return so_figs


def _get_tables_base(peeker: "Peeker", default: "DefaultReport"):
    table_divs = []
    for i, table_elem in enumerate(default.default_tables):
        table_id = peeker.id_repo.new_table_id(f"tabl-{i}")
        table_divs += [
            html.Div(id=table_id.diff, className="col-md-4"),
            *[
                html.Div(
                    table_from_pandas(table_elem.df),
                    id=id_,
                    className="col-md-4",
                    style={
                        "border-color": color,
                        "border-style": "solid",
                        "border-radius": "10px",
                        "padding": 4,
                    },
                )
                for color, id_ in zip(colors, table_id)
            ],
        ]
    return table_divs


def _get_cat_filter_elem(col, choices, selected, filter_id):

    chlist = dcc.Checklist(
        id=filter_id.id_,
        options=[{"label": v, "value": v} for v in choices],
        value=selected,
        labelStyle={"margin": 3},
        inputStyle={"margin": 3},
    )
    toggle_but = dbc.Button("Toggle", id=filter_id.toggle, className="mr-2")

    return html.Div(
        html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            html.H4(col.replace("_", " ").title()),
                            className="col-md-9",
                        ),
                        html.Div(toggle_but, className="col-md-3"),
                    ],
                    className="row",
                ),
                chlist,
                html.Hr(),
            ],
        ),
        className="col-md-12",
        style={"padding": 7},
    )


def _get_num_filter_elem(col, choices, selected, filter_id):
    n = len(choices) - 1
    slider = dcc.RangeSlider(
        id=filter_id.id_,
        min=0,
        max=n,
        value=selected or [0, n],
        marks={
            i: _format_slider_v(v)
            for i, v in enumerate(choices)
            if (i % 2 == 0) or (i == n)
        },
    )

    return html.Div(
        html.Div(
            [
                html.H4(col.replace("_", " ").title()),
                slider,
                html.Hr(),
            ],
        ),
        className="col-md-12",
        style={"padding": 3},
    )


def _format_slider_v(v):
    if v < 10:
        return "{:.2f}".format(float(v))
    return str(int(v))



from typing import TYPE_CHECKING, Iterable

from .report_classes import Metric, SharedTrace, StandaloneFigure, Table

if TYPE_CHECKING:
    from .peeker_class import Peeker  # noqa: F401


class DefaultReport:
    def __init__(self, elems, shared_figures):

        self.shared_figures = shared_figures
        self.default_elems = elems

        self._traces = {
            e.figure_name: e.trace for e in elems if isinstance(e, SharedTrace)
        }

    @property
    def default_metrics(self) -> Iterable[Metric]:
        return self._get_type_elem(Metric)

    @property
    def default_standalone_figures(self) -> Iterable[StandaloneFigure]:
        return self._get_type_elem(StandaloneFigure)

    @property
    def default_tables(self) -> Iterable[Table]:
        return self._get_type_elem(Table)

    def get_default_trace(self, figname):
        return self._traces[figname]

    def _get_type_elem(self, kls):
        return [e for e in self.default_elems if isinstance(e, kls)]



from dataclasses import dataclass, field
from typing import Union

import pandas as pd
import plotly.graph_objects as go
from plotly.basedatatypes import BaseTraceHierarchyType


@dataclass
class Metric:
    value: Union[int, str]
    name: str
    confidence_interval: list = field(default_factory=list)


@dataclass
class StandaloneFigure:
    fig: go.Figure
    col_width: int = 6


@dataclass
class SharedFigure:
    fig: go.Figure
    name: str
    col_width: int = 6


@dataclass
class SharedTrace:
    trace: BaseTraceHierarchyType
    figure_name: str


@dataclass
class Table:
    df: pd.DataFrame



import glob
import json
import os
from dataclasses import dataclass, field

DESCRIPTION_KEY = "peekview-description"


@dataclass
class PeekView:

    name: str
    description: str = ""
    filters: dict = field(default_factory=dict)


@dataclass
class ViewStore:
    root_dir: str
    _views: dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        os.makedirs(self.root_dir, exist_ok=True)
        for jspath in glob.glob(os.path.join(self.root_dir, "*.json")):
            name = os.path.split(jspath)[-1].replace(".json", "")
            with open(jspath) as fp:
                d = json.load(fp)
                desc = d.pop(DESCRIPTION_KEY, "")
                self._views[name] = PeekView(name, desc, d)

    def get(self, view_id):
        return self._views.get(view_id, PeekView(view_id))

    def set(self, view_id, desc, dic):
        d = {DESCRIPTION_KEY: desc, **dic}
        jspath = os.path.join(self.root_dir, f"{view_id}.json")
        with open(jspath, "w") as fp:
            json.dump(d, fp)
        self._views[view_id] = PeekView(view_id, desc, dic)

    def get_all_ids(self):
        return self._views.keys()



"""
TODO:
- use conf intervals of metrics (maybe groups of metrics)
- prep for gunicorn
- publish
- some pattern detection in remaining data
  - isolation forest ?
  - DBSCAN


## TESTS:

change name, add some filter, press apply button
  - updates shared fig / metric header names, updates report, updates tab label

change name, add some filter, press save button
  - updates shared fig / metric header names, updates report, updates tab label
  - updates one dd value, updates both dd options, saves json

- updates to correct new name - input vs dep

- proper update on dd select

"""

import time
from functools import reduce
from operator import and_
from typing import List, Union

import dash
import dash_html_components as html
import dask.dataframe as dd
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash.dependencies import Input, Output, State
from pyinstrument import Profiler
from structlog import get_logger

from .default_report import DefaultReport
from .elem_ids import (
    ElemIds,
    IdRepo,
    MetricId,
    SharedFigId,
    SoFigureId,
    TableId,
)
from .peeker_layout import add_base_layout_to_app
from .report_classes import (
    Metric,
    SharedFigure,
    SharedTrace,
    StandaloneFigure,
    Table,
)
from .style import LEFT_COLOR, RIGHT_COLOR, colors
from .util import (
    get_app_base,
    log_profiler,
    pandas_from_table,
    table_from_pandas,
)
from .view_store import ViewStore

logger = get_logger()


class Peeker:
    """
    API for creating and launching a peeker app

    create a subclass and properly define get_shared_figures
    and get_report_elems
    """
    def __init__(
        self,
        app_data: dd.DataFrame,
        cat_filter_cols: List[str] = None,
        num_filter_cols: List[str] = None,
        view_store_dir: str = None,
        use_profiler=False,
        range_resolution=11,
    ):
        self.ddf = app_data
        self.cat_filter_cols = cat_filter_cols or []
        self.num_filter_cols = num_filter_cols or []
        self.id_repo = IdRepo()

        self.default_values = {}
        self.num_ranges = {}
        self._set_filter_bases(range_resolution)

        self.view_store = ViewStore(view_store_dir or "./view_store_data")

        self.app = get_app_base(__name__)

        add_base_layout_to_app(
            self.app,
            self,
            DefaultReport(
                self.get_report_elems(self.ddf) or [],
                self.get_shared_figures() or [],
            ),
        )

        self._add_callbacks()

        self._timing = use_profiler

    def run(self, port_no=5666, host="0.0.0.0", debug=True):
        self.app.run_server(host=host, port=port_no, debug=debug)

    def get_shared_figures(self) -> List[SharedFigure]:
        """overwrite this function

        returns a list of figures that provides a frame for
        the shared traces from get_anal_elems
        """

        return [
            SharedFigure(go.Figure(layout=go.Layout(title=c.title())), c)
            for c in self.cat_filter_cols
        ]

    def get_report_elems(
        self, filtered_ddf: dd.DataFrame
    ) -> List[Union[Metric, SharedTrace, StandaloneFigure, Table]]:
        """overwrite this function

        returns analysis elements
        """
        elems = [
            Metric(filtered_ddf.shape[0].compute(), "Size"),
            Table(filtered_ddf.head(3).iloc[:, :3]),
        ]

        if self.num_filter_cols:
            c = self.num_filter_cols[0]
            elems.append(
                Metric(filtered_ddf.loc[:, c].mean().compute(), f"Average {c}")
            )

        for c in self.cat_filter_cols:
            s = filtered_ddf.loc[:, c].value_counts().compute()
            elems.append(
                SharedTrace(go.Bar(x=s.index.tolist(), y=s.tolist()), c)
            )

        if self.cat_filter_cols and self.num_filter_cols:
            c = self.cat_filter_cols[0]
            nc = self.num_filter_cols[0]
            s = filtered_ddf.groupby(c)[nc].mean().compute()
            elems.append(
                StandaloneFigure(
                    go.Figure(data=[go.Bar(x=s.index.tolist(), y=s.tolist())])
                )
            )

        return elems

    def _set_filter_bases(self, range_resolution):
        for col in self.cat_filter_cols:
            self.id_repo.new_filter_id(col)
            self.default_values[col] = sorted(
                self.ddf.loc[:, col].drop_duplicates().compute().tolist()
            )

        for col in self.num_filter_cols:
            self.id_repo.new_filter_id(col)
            binends = (
                self.ddf.loc[:, col]
                .quantile(np.linspace(0, 1, range_resolution))
                .compute()
                .drop_duplicates()
                .tolist()
            )
            selected = [0, len(binends) - 1]
            self.default_values[col] = selected
            self.num_ranges[col] = binends

    def _add_callbacks(self):

        self._add_saving_callbacks()
        self._add_report_callbacks()
        self._add_dd_callbacks()
        self._add_metric_diff_callbacks()
        self._add_shared_trace_callbacks()
        self._add_table_diff_callbacks()

    def _add_saving_callbacks(self):

        for ind, side_id in enumerate(ElemIds.sides()):
            self.app.callback(
                Output(side_id.dd, "value"),
                Output(side_id.saved_view_option_store, "data"),
                Input(side_id.save_button, "n_clicks"),
                State(side_id.name_input, "value"),
                State(side_id.desc_input, "value"),
                *self._get_filter_states(ind),
                prevent_initial_call=True,
            )(self._save_side)

        @self.app.callback(
            Output(ElemIds.left.dd, "options"),
            Output(ElemIds.right.dd, "options"),
            Input(ElemIds.left.saved_view_option_store, "data"),
            Input(ElemIds.right.saved_view_option_store, "data"),
            prevent_initial_call=True,
        )
        def add_option(_, __):
            options = [
                {"value": x, "label": x} for x in self.view_store.get_all_ids()
            ]
            return options, options

    def _add_report_callbacks(self):

        for ind, side_id in enumerate(ElemIds.sides()):
            self.app.callback(
                Output(side_id.metric_table_header, "children"),
                Output(side_id.tab, "label"),
                *self._get_report_output_dependencies(ind),
                Input(side_id.apply_button, "n_clicks"),
                State(side_id.dd, "value"),
                State(side_id.name_input, "value"),
                *self._get_filter_states(ind),
                prevent_initial_call=True,
            )(self._report_from_filters)

    def _add_dd_callbacks(self):
        for ind, side_id in enumerate(ElemIds.sides()):

            @self.app.callback(
                Output(side_id.apply_button, "n_clicks"),
                Output(side_id.name_input, "value"),
                Output(side_id.desc_input, "value"),
                *[
                    Output(fid.id_, "value")
                    for fid in self._get_filter_ids(ind)
                ],
                Input(side_id.dd, "value"),
                *[
                    Input(fid.toggle, "n_clicks")
                    for fid in self._get_cat_filter_ids(ind)
                ],
                State(side_id.name_input, "value"),
                State(side_id.desc_input, "value"),
                *[self._get_filter_states(ind)],
                prevent_initial_call=True,
            )
            def update_vals(view_id, *toggles_values):

                name = toggles_values[self._n_cat_filters]
                desc = toggles_values[self._n_cat_filters + 1]

                # toggles = toggles_values[: self._n_cat_filters]
                _vindl = self._n_cat_filters + 2
                values = list(toggles_values[_vindl:])
                n_clicks = 0

                for elem in dash.callback_context.triggered:
                    if ElemIds.is_dd(elem["prop_id"]):
                        n_clicks = 1
                        peek_view = self.view_store.get(
                            view_id or "Unnamed View"
                        )
                        name = peek_view.name
                        desc = peek_view.description
                        for i, fid in enumerate(self._get_filter_ids(ind)):
                            col = fid.name
                            values[i] = peek_view.filters.get(
                                col, self.default_values[col]
                            )
                    if elem["prop_id"].endswith(".n_clicks"):
                        nclicks = elem["value"]
                        col, f_ind = self._get_toggle_col_and_ind(
                            elem["prop_id"]
                        )
                        if nclicks % 2 == 0:
                            nuvals = self.default_values[col]
                        else:
                            nuvals = []
                        values[f_ind] = nuvals

                return (n_clicks, name, desc, *values)

    def _add_metric_diff_callbacks(self):
        for mid in self._get_metric_ids():

            @self.app.callback(
                Output(mid.diff, "children"),
                Output(mid.diff, "style"),
                Input(mid.left, "children"),
                Input(mid.right, "children"),
            )
            def update_mdiff(val1, val2):
                nv1 = float(val1 or "nan")
                nv2 = float(val2 or "nan")
                m = (abs(nv1) + abs(nv2)) / 2

                color = colors[int(nv1 < nv2)]

                diff = "+ {:.02f}%".format(abs(nv1 - nv2) * 100 / m)

                return diff, {"color": color}

    def _add_shared_trace_callbacks(self):
        for shid in self._get_shared_report_ids():

            @self.app.callback(
                Output(shid.id_, "figure"),
                Input(shid.left, "data"),
                Input(shid.right, "data"),
                State(shid.id_, "figure"),
                prevent_initial_call=True,
            )
            def update_shared_fig(trace1, trace2, fig):
                for trace, ntrace in zip(fig["data"], [trace1, trace2]):
                    trace.update(ntrace)
                return fig

    def _add_table_diff_callbacks(self):
        for tabid in self._get_table_ids():

            @self.app.callback(
                Output(tabid.diff, "children"),
                Input(tabid.left, "children"),
                Input(tabid.right, "children"),
            )
            def update_tabdiff(tab1, tab2):
                df1 = pandas_from_table(tab1)
                dfi1 = df1.iterrows()
                dfi2 = pandas_from_table(tab2).iterrows()

                recs = []
                for (i1, r1), (i2, r2) in zip(dfi1, dfi2):
                    recs.append(
                        [
                            html.B(e1)
                            if (e1 == e2)
                            else [
                                html.Span(e1, style={"color": LEFT_COLOR}),
                                html.B(" vs "),
                                html.Span(e2, style={"color": RIGHT_COLOR}),
                            ]
                            for e1, e2 in zip(r1, r2)
                        ]
                    )

                return table_from_pandas(
                    pd.DataFrame(recs, columns=df1.columns)
                )

    def _save_side(
        self,
        _,
        name,
        desc,
        *side_filter_vals,
    ):

        d = {}
        for fid, val in zip(self._get_filter_ids(), side_filter_vals):
            d[fid.name] = val
        self.view_store.set(name, desc, d)

        return name, list(self.view_store.get_all_ids())

    def _report_from_filters(
        self, n_clicks, dd_name, input_name, *filter_values
    ):

        if n_clicks == 0:
            raise ValueError("no true report request, only toggle")

        stime = time.time()
        name = input_name
        for elem in dash.callback_context.triggered:
            if ElemIds.is_dd(elem["prop_id"]):
                name = dd_name

        logger.info("start filtering")
        if self._timing:
            logger.info("starting time profiling")
            profiler = Profiler()
            profiler.start()

        filtered_ddf = self._filter_ddf(filter_values)

        logger.info("calculating elements")
        elems = self.get_report_elems(filtered_ddf)
        logger.info(f"{len(elems)} elements returned")

        logger.info("parsing elements to output values")
        outputs = self._parse_report_elems_to_outputs(elems, name)
        logger.info(f"{len(outputs)} outputs parsed")
        logger.info("report done", runtime=round(time.time() - stime, 3))

        if self._timing:
            profiler.stop()
            log_profiler(profiler, logger, 20)

        return (name, name, *outputs)

    def _filter_ddf(self, filter_values):
        filt_series = []

        for fval, col in zip(filter_values, self._filter_cols):
            s = self.ddf.loc[:, col]
            if col in self.cat_filter_cols:
                new_filt_series = s.isin(fval)
            else:
                num_range = self.num_ranges[col]
                vmin = num_range[fval[0]]
                vmax = num_range[fval[1]]
                new_filt_series = (s >= vmin) & (s <= vmax)

            filt_series.append(new_filt_series)

        return self.ddf.loc[reduce(and_, filt_series), :]

    def _parse_report_elems_to_outputs(self, elems, name):

        metric_out = []
        shared_out = []
        so_out = []
        table_out = []

        for elem in elems:
            if isinstance(elem, Metric):
                # TODO conf interval here
                metric_out.append(elem.value)
            if isinstance(elem, SharedTrace):
                trace = elem.trace
                trace.name = name
                shared_out.append(trace)
            if isinstance(elem, StandaloneFigure):
                so_out.append(elem.fig)
            if isinstance(elem, Table):
                table_out.append(table_from_pandas(elem.df))
        return [*metric_out, *shared_out, *so_out, *table_out]

    def _get_filter_ids(self, ind=0):
        return self.id_repo.filter_ids[ind]

    def _get_cat_filter_ids(self, ind=0):
        return [
            fid
            for fid in self.id_repo.filter_ids[ind]
            if fid.name in self.cat_filter_cols
        ]

    def _get_filter_states(self, ind):
        return [State(fid.id_, "value") for fid in self._get_filter_ids(ind)]

    def _get_toggle_col_and_ind(self, prop_id):
        for i, (fidl, fidr) in enumerate(
            zip(*self.id_repo.filter_ids.values())
        ):
            if prop_id.startswith(fidl.toggle) or prop_id.startswith(
                fidr.toggle
            ):
                return fidr.name, i

    def _get_shared_report_ids(self):
        return self.id_repo.shared_fig_ids

    def _get_metric_ids(self):
        return self.id_repo.metric_ids

    def _get_table_ids(self):
        return self.id_repo.table_ids

    def _get_report_output_dependencies(self, ind):

        outputs = []
        for eid in self.id_repo.report_elem_ids:
            if isinstance(eid, (MetricId, TableId)):
                outputs.append(Output(eid[ind], "children"))
            if isinstance(eid, SharedFigId):
                outputs.append(Output(eid[ind], "data"))
            if isinstance(eid, SoFigureId):
                outputs.append(Output(eid[ind], "figure"))
        return outputs

    @property
    def _n_shared_states(self):
        return len(self.id_repo.shared_fig_ids)

    @property
    def _n_cat_filters(self):
        return len(self.cat_filter_cols)

    @property
    def _filter_cols(self):
        return [*self.cat_filter_cols, *self.num_filter_cols]



from ._version import __version__  # noqa: F401
from .peeker_class import Peeker  # noqa: F401
from .report_classes import (  # noqa: F401
    Metric,
    SharedFigure,
    SharedTrace,
    StandaloneFigure,
    Table,
)



import tempfile
import time
from multiprocessing import Process
from typing import Optional

import dask.dataframe as dd
import requests

from daskpeeker import Peeker
from daskpeeker.tests.data_for_testing import CATEG_COLS, NUM_COLS, get_test_df

# from dask.distributed import Client, get_client


test_app_default_port = 5677
test_peeker_default_address = f"http://localhost:{test_app_default_port}"


class PeekerRunner:
    def __init__(self, port_no=test_app_default_port, verbose=True):
        self._tmpdir = tempfile.TemporaryDirectory()
        self._port_no = port_no
        self.app_address = f"http://localhost:{port_no}"
        self._process: Optional[Process] = None
        self._verbose = verbose
        # try:
        #    self.client = get_client()
        # except ValueError as e:
        #    print("new client started: ", e)
        #    self.client = Client()
        nparts = 4  # sum(self.client.nthreads().values()

        test_ddf = dd.from_pandas(get_test_df(), npartitions=nparts).persist()

        self.peeker = Peeker(test_ddf, CATEG_COLS, NUM_COLS, self._tmpdir.name)

    def start(self):
        self._process = Process(
            target=self.peeker.app.run_server, kwargs={"port": self._port_no}
        )
        self._process.start()
        while True:
            try:
                requests.get(self.app_address, timeout=5)
                break
            except requests.exceptions.ConnectionError:
                pass
            time.sleep(0.2)

    def stop(self):
        # self.client.shutdown()
        # self.client.close()
        self._process.kill()
        self._tmpdir.cleanup()



import random

import faker
import numpy as np
import pandas as pd

CATEG_COLS = ["c1", "c2"]
NUM_COLS = ["n1", "n2", "nlog", "nlog2"]


def get_test_df(n=500, seed=42069):

    fake = faker.Faker()
    np.random.seed(seed)
    faker.Faker.seed(seed)

    cs1 = [fake.name() for _ in range(7)]
    cs2 = [np.random.randint(10000) for _ in range(10)]

    return pd.DataFrame(
        {
            CATEG_COLS[0]: random.choices(cs1, k=n),
            CATEG_COLS[1]: random.choices(cs2, k=n),
            NUM_COLS[0]: np.random.rand(n),
            NUM_COLS[1]: np.random.randint(10000, size=n),
            NUM_COLS[2]: np.random.lognormal(8, 3, size=n),
            NUM_COLS[3]: np.random.lognormal(0, 10, size=n).astype(int),
        }
    )



from typing import Iterable


class _IdBase:
    def __init__(self, name, suffix=None):
        self.name = name
        self.id_ = (
            name.replace(" ", "-") + "-" + (suffix or type(self).__name__)
        ).lower()


class _ReportElemIdBase(_IdBase):
    @property
    def left(self):
        return f"{self.id_}-left"

    @property
    def right(self):
        return f"{self.id_}-right"

    @property
    def diff(self):
        return f"{self.id_}-diff"

    def __iter__(self):
        yield self.left
        yield self.right

    def __getitem__(self, key):
        return [self.left, self.right][key]


class MetricId(_ReportElemIdBase):
    pass


class SharedFigId(_ReportElemIdBase):
    pass


class SoFigureId(_ReportElemIdBase):
    pass


class TableId(_ReportElemIdBase):
    pass


class FilterId(_IdBase):
    @property
    def toggle(self):
        return f"{self.id_}-toggle"


class IdRepo:
    def __init__(self):
        self.report_elem_ids = []
        self.filter_ids = {0: [], 1: []}

    def new_metric_id(self, metric_name):
        return self._new_id_type(metric_name, MetricId)

    def new_shared_fig_id(self, fig_name):
        return self._new_id_type(fig_name, SharedFigId)

    def new_so_fig_id(self, fig_name):
        return self._new_id_type(fig_name, SoFigureId)

    def new_table_id(self, name):
        return self._new_id_type(name, TableId)

    def new_filter_id(self, name):
        self.filter_ids[0].append(FilterId(name, "left"))
        self.filter_ids[1].append(FilterId(name, "right"))

    def get_filter_id(self, col, ind):
        for fid in self.filter_ids[ind]:
            if fid.name == col:
                return fid

    @property
    def shared_fig_ids(self):
        return [
            eid for eid in self.report_elem_ids if isinstance(eid, SharedFigId)
        ]

    @property
    def metric_ids(self):
        return [
            eid for eid in self.report_elem_ids if isinstance(eid, MetricId)
        ]

    @property
    def table_ids(self):
        return [
            eid for eid in self.report_elem_ids if isinstance(eid, TableId)
        ]

    def _new_id_type(self, name, kls):
        new_id = kls(name)
        self.report_elem_ids.append(new_id)
        return new_id


class SidedId:
    def __init__(self, side):
        self.dd = f"dd-{side}"
        self.tab = f"tab-{side}"

        self.name_input = f"name-{side}"
        self.desc_input = f"desc-{side}"

        self.save_button = f"save-button-{side}"
        self.del_button = f"del-button-{side}"
        self.apply_button = f"apply-button-{side}"

        self.saved_view_option_store = f"option-store-{side}"

        self.metric_table_header = f"metric-table-header-{side}"


class ElemIds:

    navbar = "navbar"
    sidebar = "sidebar"
    main_content = "main-content"

    sidebar_button = "sidebar-toggle"
    url = "url"
    click_store = "click-store"

    left = SidedId("left")
    right = SidedId("right")

    @classmethod
    def sides(cls) -> Iterable[SidedId]:
        return [cls.left, cls.right]

    @classmethod
    def is_dd(cls, prop_id):
        return SidedId("").dd in prop_id



DEFAULT_VIEW_NAME = "Unnamed View"

LEFT_COLOR = "#00AEF9"
RIGHT_COLOR = "#F9AE00"
colors = [LEFT_COLOR, RIGHT_COLOR]

sidebar_w = 45
content_l_m = 0
navb_h = 50  # 62.5

SIDEBAR_STYLE = {
    "position": "fixed",
    "top": navb_h,
    "left": 0,
    "bottom": 0,
    "width": f"{sidebar_w}rem",
    "height": "100%",
    "z-index": 1,
    "overflow-x": "hidden",
    "transition": "all 0.5s",
    # "padding": "0.5rem 1rem",
    "background-color": "#f8f9fa",
}

SIDEBAR_HIDDEN = {
    **SIDEBAR_STYLE,
    "left": f"-{sidebar_w}rem",
    "padding": "0rem 0rem",
}


CONTENT_STYLE = {
    "transition": "margin-left .5s",
    "margin-left": f"{content_l_m + sidebar_w}rem",
    # "margin-right": "2rem",
    "padding": "4rem",
    "background-color": "#f8f9fa",
}

CONTENT_STYLE_HIDDEN = {
    **CONTENT_STYLE,
    "margin-left": f"{content_l_m}rem",
}



import requests
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from daskpeeker.tests.peeker_for_testing import test_peeker_default_address

left_c1_input_1_xpath = '//*[@id="c1-left"]/label[1]/input'
left_apply_xpath = '//*[@id="apply-button-left"]'


def _init_browser(browser):
    browser.get(test_peeker_default_address)
    WebDriverWait(browser, 10).until(
        EC.presence_of_element_located((By.XPATH, left_c1_input_1_xpath))
    )


def test_peeker_setup(browser):
    requests.get(test_peeker_default_address)
    _init_browser(browser)

    elem = browser.find_element_by_xpath(left_c1_input_1_xpath)
    elem.click()

    filt_but = browser.find_element_by_xpath(left_apply_xpath)
    filt_but.click()



from functools import reduce
from operator import and_

import dash
import dash_html_components as html
import pandas as pd

bt_root = "https://maxcdn.bootstrapcdn.com/bootstrap/4.5.2/"

external_stylesheets = [
    "https://codepen.io/chriddyp/pen/bWLwgP.css",
    f"{bt_root}css/bootstrap.min.css",
    f"{bt_root}css/bootstrap-theme.min.css",
    "https://use.fontawesome.com/releases/v5.7.2/css/all.css",
]

external_scripts = [f"{bt_root}js/bootstrap.min.js"]


def get_app_base(name=__name__, suppress_cbe=False):
    return dash.Dash(
        name,
        external_stylesheets=external_stylesheets,
        external_scripts=external_scripts,
        suppress_callback_exceptions=suppress_cbe,
    )


def table_from_pandas(df):
    index_name = df.index.name
    headers = []
    if index_name:
        headers.append(index_name)
    rows = [html.Tr([html.Th(c) for c in [*headers, *df.columns]])]

    for ind, r in df.iterrows():
        elems = []
        if index_name:
            elems.append(ind)
        rows.append(html.Tr([html.Td(e) for e in [*elems, *r]]))
    return html.Table(rows)


def pandas_from_table(tabdic):
    rows = tabdic["props"]["children"]
    colnames = [td["props"]["children"] for td in rows[0]["props"]["children"]]
    records = [
        [td["props"]["children"] for td in row["props"]["children"]]
        for row in rows[1:]
    ]

    return pd.DataFrame(records, columns=colnames)


def get_profile_df(profiler, topn=15):
    to_drop = [
        "tornado",
        "IPython",
        "ipython",
        "ipykernel",
        "runpy.py",
        "asyncio",
        "werkzeug/",
        "flask/",
        "dash/",
        "socketserver.py",
        "threading.py",
        "http/server",
        "importlib",
        "dask/",
    ]
    return (
        pd.DataFrame(
            [
                {
                    "time": r[1],
                    **dict(zip(["function", "file", "line"], p.split("\x00"))),
                }
                for r in profiler.frame_records
                for p in r[0]
            ]
        )
        .loc[
            lambda df: reduce(
                and_, [~df["file"].str.contains(s) for s in to_drop]
            )
        ]
        .assign(
            pend=lambda df: df["file"].str.split("/").str[-4:].str.join("/")
        )
        .groupby(["pend", "line", "function"])
        .sum()
        .nlargest(topn, "time")
    )


def log_profiler(profiler, logger, topn=15):
    for (path, line_no, function), r in get_profile_df(
        profiler, topn
    ).iterrows():
        logger.info(
            "profiled period top",
            duration=r["time"],
            function=function,
            path=path,
            line_no=line_no,
        )
